// MainFrm.h
#pragma once

#include "FactoryPane.h"
#include "SidePane.h"
#include "PaneSplitter.h"

#include "NewDialog.h"
#include "OpenDialog.h"
#include "OptionProp.h"

#include "ReBar.h"
#include "MenuBar.h"
#include "ToolBar.h"
#include "AddressBar.h"
#include "StatusBar.h"

const DWORD MSG_MAP_ID_GESTURE	= 1;

namespace Default
{
	const int	SideBarWidth		= 200;
	const bool	SplitVertical		= true;
	const int	AddressBarHeight	= 300;
	const int	TabWidthMin			=  24;
	const int	TabWidthMax			= 200;
	const bool	EnableXPMenu		= false;
}

const WORD IDW_REBAR		= 100;
const WORD IDW_MENUBAR		= 201;
const WORD IDW_TOOLBAR		= 202;
const WORD IDW_ADDRESSBAR	= 203;

// 
// EBhEKw\
// - CMainFrame
//   - m_FactoryPane (invisible)
//   - m_ReBar
//     - m_MenuBar
//     - m_ToolBar
//     - m_AddressBar
//   - m_StatusBar
//   - m_RootSplitter
//     - m_SidePane
//     - m_PaneSplitter
//       - CViewPane array
//

class CMainFrame;
typedef CMouseGestureImpl<CMainFrame, MSG_MAP_ID_GESTURE>	CMainFrame_Gesture;
typedef CStringTableMenuHandler< CMenuHandler<CMainFrame> >	CMainFrame_MenuHandler;
typedef CMenuProvider<CMainFrame>							CMainFrame_MenuProvider;

class CMainFrame :
	public ITablet,
	public CSimpleFrameImpl<CMainFrame>,
	public CUpdateUI<CMainFrame>,
	public CMessageFilter,
	public CIdleHandler,
	public CDropTargetImpl<CMainFrame>,
	public CMouseWheelHandler<CMainFrame>,
	public CMainFrame_Gesture,
	public CMainFrame_MenuHandler,
	public CMainFrame_MenuProvider
{
	typedef CSimpleFrameImpl<CMainFrame>	baseClass;
	typedef CUpdateUI<CMainFrame>			baseUpdateUI;
	typedef CMouseWheelHandler<CMainFrame>	baseWheel;
	typedef CMainFrame_Gesture				baseGesture;
	typedef CMainFrame_MenuProvider			baseMenuProvider;
	typedef CMainFrame_MenuHandler			baseMenuHandler;

private:
	// Main
	CFactoryPane	m_FactoryPane;
	CSplitterWindow	m_RootSplitter; // == m_hWndClient
	CSidePane		m_SidePane;
	CPaneSplitter	m_PaneSplitter;

	// Tools
	CReBar			m_ReBar; // == m_hWndToolBar
	CMenuBar		m_MenuBar;
	CToolBar		m_ToolBar;
	CAddressBar		m_AddressBar;
	CStatusBar		m_StatusBar; // == m_hWndStatusBar

	// Dialogs
	CNewDialog		m_NewDlg;
	COpenDialog		m_OpenDlg;

	CStdVector<HINSTANCE> m_Libs;

public:
	DECLARE_FRAME_WND_CLASS_EX(_T("Tablet"), IDR_MAINFRAME, 0, -1)

	CMainFrame()
	{
		ATLASSERT(theTablet == NULL);
		theTablet = this;
	}
	~CMainFrame()
	{
		ATLASSERT(theTablet == this);
		theTablet = NULL;
	}

public:
	void LoadProfile(CProfile& profile)
	{
		const CReBarBandInfo bands[] = 
		{
			CReBarBandInfo(m_MenuBar	, NULL),
			CReBarBandInfo(m_ToolBar	, NULL),
			CReBarBandInfo(m_AddressBar	, _T("AhX")),
		};
		profile.LoadWindowPlacement(m_hWnd);
		profile.LoadReBarState(m_ReBar, bands, sizeof(bands)/sizeof(bands[0]));
		EnableTool(WINDOW_STATUSBAR, profile.ReadBool(_T("StatusBar"), true));
		EnableTool(WINDOW_SIDEBAR  , profile.ReadBool(_T("SideBar")  , true));
		UpdateLayout(TRUE);
		m_RootSplitter.SetSplitterPos(profile.ReadInt(_T("SideBarWidth"), Default::SideBarWidth));
		m_PaneSplitter.SetVertical(profile.ReadBool(_T("SplitVertical"), Default::SplitVertical));
	}
	void SaveProfile(CProfile& profile)
	{
		profile.SaveWindowPlacement(m_hWnd);
		profile.SaveReBarState(m_ReBar);
		profile.WriteBool(_T("StatusBar"), IsToolEnabled(WINDOW_STATUSBAR));
		profile.WriteBool(_T("SideBar")  , IsToolEnabled(WINDOW_SIDEBAR));
		profile.WriteInt (_T("SideBarWidth"), m_RootSplitter.GetSplitterPos());
		profile.WriteBool(_T("SplitVertical"), m_PaneSplitter.IsVertical());
	}

public:
	////////////////////////////////////////////////////////////////////////////////
	// ITablet

	HWND GetWindow(WindowID id)
	{
		switch(id)
		{
		case WINDOW_MAINFRAME:
			ATLASSERT(IsWindow());
			return m_hWnd;
		case WINDOW_FACTORY:
			ATLASSERT(m_FactoryPane.IsWindow());
			return m_FactoryPane;
		case WINDOW_MENUBAR:
			return m_MenuBar;
		case WINDOW_TOOLBAR:
			return m_ToolBar;
		case WINDOW_ADDRESSBAR:
			return m_AddressBar;
		case WINDOW_STATUSBAR:
			ATLASSERT(m_StatusBar.IsWindow());
			return m_StatusBar;
		case WINDOW_SIDEBAR:
			ATLASSERT(m_SidePane.IsWindow());
			return m_SidePane;
		case WINDOW_PANESPLITTER:
			ATLASSERT(m_PaneSplitter.IsWindow());
			return m_PaneSplitter;
		case WINDOW_VIEW_ACTIVE:
			return m_PaneSplitter.GetActiveElement().GetActiveElement();
		case WINDOW_PANE_ACTIVE:
			return m_PaneSplitter.GetActiveElement();
		default:
			if(WINDOW_PANE_FIRST <= id && id <= WINDOW_PANE_LAST)
			{
				int nPane = id - WINDOW_PANE_FIRST;
				if(m_PaneSplitter.EnsurePaneCount(nPane+1))
					return m_PaneSplitter.GetElement(nPane);
				else
					return NULL;
			}
			else
			{
				ATLTRACE("ERROR: Invalid WindowID. ITablet::GetWindow(%d)\n", id);
				return NULL;
			}
		}
	}

	UINT CountElements(WindowID id)
	{
		switch(id)
		{
		case WINDOW_FACTORY:
			if(!m_FactoryPane.IsWindow())
				return 0;
			return m_FactoryPane.CountElements();
		case WINDOW_SIDEBAR:
			if(!m_SidePane.IsWindow())
				return 0;
			return m_SidePane.CountElements();
		case WINDOW_PANESPLITTER:
			if(!m_PaneSplitter.IsWindow())
				return 0;
			return m_PaneSplitter.CountElements();
		case WINDOW_PANE_ACTIVE:
			return m_PaneSplitter.GetActiveElement().CountElements();
		default:
			if(WINDOW_PANE_FIRST <= id && id <= WINDOW_PANE_LAST)
			{
				int nPane = id - WINDOW_PANE_FIRST;
				if(m_PaneSplitter.EnsurePaneCount(nPane+1))
					return m_PaneSplitter.GetElement(nPane).CountElements();
				else
					return 0;
			}
			else
			{
				ATLTRACE("ERROR: Invalid WindowID. ITablet::CountElements(%d)\n", id);
				return NULL;
			}
		}
	}

	HWND GetElement(WindowID id, UINT index)
	{
		switch(id)
		{
		case WINDOW_FACTORY:
			if(!m_FactoryPane.IsWindow())
				return NULL;
			return m_FactoryPane.GetElement(index);
		case WINDOW_SIDEBAR:
			if(!m_SidePane.IsWindow())
				return NULL;
			return m_SidePane.GetElement(index);
		case WINDOW_PANESPLITTER:
			return m_PaneSplitter.GetElement(index);
		case WINDOW_PANE_ACTIVE:
			return m_PaneSplitter.GetActiveElement().GetElement(index);
		default:
			if(WINDOW_PANE_FIRST <= id && id <= WINDOW_PANE_LAST)
			{
				UINT nPane = id - WINDOW_PANE_FIRST;
				m_PaneSplitter.EnsurePaneCount(nPane+1);
				return m_PaneSplitter.GetElement(nPane).GetElement(index);
			}
			else
			{
				ATLTRACE("ERROR: Invalid WindowID. ITablet::GetElement(%d)\n", id);
				return NULL;
			}
		}
	}

	BOOL MoveElement(HWND hWnd, UINT pane, UINT index)
	{
		UINT nPaneCount = CountElements(WINDOW_PANESPLITTER);
		for(UINT i = 0; i < nPaneCount; i++)
		{
			CPaneWindow prev = m_PaneSplitter.GetElement(i);
			UINT nViewCount = prev.CountElements();
			for(UINT j = 0; j < nViewCount; j++)
			{
				if(hWnd == prev.GetElement(j))
				{
					if(pane > nPaneCount)  pane = nPaneCount;
					if(!m_PaneSplitter.EnsurePaneCount(pane+1))
						return false;
					CPaneWindow next = m_PaneSplitter.GetElement(pane);
					ATLTRACE("MoveElement HWND=%08X, (%d, %d) => (%d, %d)\n", hWnd, i, j, pane, index);
					if(pane != i)
					{
						return prev.DeleteElement(j) && next.AddElement(hWnd, index);
					}
					else if(index != j)
					{
						return next.MoveElement(hWnd, index);
					}
				}
			}
		}
		return FALSE;
	}

	BOOL OpenFile(const FileEntry& entry, WindowID pane, UINT document, INT minSuitLevel)
	{
		switch(entry.type)
		{
		case FileEntry::TYPE_PATHA:
			return DoOpenFile(NULL, CA2CT(entry.pPathA), pane, document, minSuitLevel);
		case FileEntry::TYPE_PATHW:
			return DoOpenFile(NULL, CW2CT(entry.pPathW), pane, document, minSuitLevel);
		case FileEntry::TYPE_ITEMIDLIST:
			return DoOpenFile((const CShellItem*)entry.pIDL, NULL, pane, document, minSuitLevel);
		case FileEntry::TYPE_CIDA:
		{
			const CShellItemList* pItems = static_cast<const CShellItemList*>(entry.pCIDA);
			UINT cFiles = pItems->GetCount();
			for(UINT i = 0; i < cFiles; i++)
			{
				CAutoPtr<CShellItemTask> pFullPath = CShellItemTask::Concat(pItems->GetParent(), pItems->GetAt(i));
				DoOpenFile(pFullPath, NULL, pane, document, minSuitLevel);
			}
			return TRUE;
		}
		case FileEntry::TYPE_HDROP:
		{
			UINT cFiles = ::DragQueryFile(entry.hDrop, 0xFFFFFFFF, 0, 0);
			for(UINT i = 0; i < cFiles; i++)
			{
				TCHAR path[MAX_PATH];
				if(::DragQueryFile(entry.hDrop, i, path, MAX_PATH) > 0)
					DoOpenFile(NULL, path, pane, document, minSuitLevel);
			}
			return TRUE;
		}
		default:
			ATLTRACE("Invalid FileEntry.type == %d\n", entry.type);
			return FALSE;
		}
	}

	DWORD GetControlStyle(ControlType type)
	{
		switch(type)
		{
		case CONTROL_LISTVIEW:
			return 0;
		case CONTROL_TREEVIEW:
			//return TVS_FULLROWSELECT | TVS_HASBUTTONS | TVS_SHOWSELALWAYS | TVS_SINGLEEXPAND | TVS_TRACKSELECT;
			return TVS_HASLINES | TVS_HASBUTTONS | TVS_SHOWSELALWAYS;
		default:
			return 0;
		}
	}

	HMENU LoadMenu(VARIANT xml)
	{
		return NULL;
	}

	HACCEL LoadAccelerators(VARIANT xml)
	{
		return NULL;
	}

	HIMAGELIST GetImageList(bool big)
	{
		return NULL;
	}

	UINT CountHistory()
	{
		return m_AddressBar.GetCount();
	}

	LPCTSTR GetHistory(UINT index)
	{
		return m_AddressBar.GetAt(index);
	}

	UINT AddHistory(LPCTSTR path)
	{
		return m_AddressBar.Add(path);
	}

	UINT TrackPopupMenu(HMENU hMenu, POINT pt, HWND hWnd, UINT uFlags, LPTPMPARAMS lpParams)
	{
		return baseMenuProvider::TrackPopupMenu(hMenu, pt, hWnd, uFlags, lpParams);
	}

	UINT DoTrackPopupMenu(HMENU hMenu, POINT pt, HWND hWnd, UINT uFlags, LPTPMPARAMS lpParams)
	{
		// m_MenuBar XPj[voC_ƂėpB
		return m_MenuBar.TrackPopupMenu(hMenu, uFlags, pt.x, pt.y, lpParams);
	}

	bool IsToolEnabled(WindowID tool) const
	{
		switch(tool)
		{
		case WINDOW_MENUBAR:
			return m_ReBar.IsBandVisible(IDW_MENUBAR);
		case WINDOW_TOOLBAR:
			return m_ReBar.IsBandVisible(IDW_TOOLBAR);
		case WINDOW_ADDRESSBAR:
			return m_ReBar.IsBandVisible(IDW_ADDRESSBAR);
		case WINDOW_STATUSBAR:
			return m_StatusBar.IsWindowEnabled() != 0;
		case WINDOW_SIDEBAR:
			return m_RootSplitter.GetSinglePaneMode() == SPLIT_PANE_NONE;
		default:
			ATLTRACE("Invalid WindowID : IsToolEnabled(%d)\n", tool);
			return false;
		}
	}

	bool EnableTool(WindowID tool, bool enable)
	{
		switch(tool)
		{
		case WINDOW_MENUBAR:
			m_ReBar.SetBandVisible(IDW_MENUBAR, enable);
			UISetCheck(ID_VIEW_MENUBAR, enable);
			break;
		case WINDOW_TOOLBAR:
			m_ReBar.SetBandVisible(IDW_TOOLBAR, enable);
			UISetCheck(ID_VIEW_TOOLBAR, enable);
			break;
		case WINDOW_ADDRESSBAR:
			m_ReBar.SetBandVisible(IDW_ADDRESSBAR, enable);
			UISetCheck(ID_VIEW_ADDRESSBAR, enable);
			break;
		case WINDOW_STATUSBAR:
			m_StatusBar.EnableWindow(enable);
			m_StatusBar.ShowWindow(enable ? SW_SHOWNOACTIVATE : SW_HIDE);
			UISetCheck(ID_VIEW_STATUS_BAR, enable);
			break;
		case WINDOW_SIDEBAR:
			m_SidePane.EnableWindow(enable);
			m_SidePane.ShowWindow(enable ? SW_SHOW : SW_HIDE);
			m_RootSplitter.SetSinglePaneMode(enable ? SPLIT_PANE_NONE : SPLIT_PANE_RIGHT);
			UISetCheck(ID_VIEW_SIDEBAR, enable);
			break;
		default:
			ATLTRACE("Invalid WindowID : EnableTool(%d)\n", tool);
			return false;
		}
		UpdateLayout();
		return enable;
	}

	bool ToggleToolEnabled(WindowID tool)
	{
		return EnableTool(tool, !IsToolEnabled(tool));
	}

	STDMETHODIMP GetControlWindow(UINT id, HWND * lphwnd)
	{
		if(lphwnd == NULL)
			return E_POINTER;
		if(FCW_STATUS == id)
		{
			*lphwnd = m_hWndStatusBar;
			return S_OK;
		}
		return E_NOTIMPL;
	}

private:
	////////////////////////////////////////////////////////////////////////////////
	//

	BOOL DoOpenFile(const CShellItem* pItem, LPCTSTR filename, WindowID pane, UINT document, INT minSuitLevel)
	{
		if(pane < WINDOW_PANE_ACTIVE || WINDOW_PANE_LAST < pane)
		{
			ATLTRACE(_T("ERROR: DoOpenFile(%d) sȃyC\n"), pane);
			return FALSE;
		}
		// pItem ܂ filename  NULL̏ꍇɎgpobt@B
		CAutoPtr<CShellItemTask> pItemBuffer;
		TCHAR pathBuffer[MAX_PATH];
		//
		CShellLink link;
		HRESULT hr = E_FAIL;
		if(pItem)
			hr = link.SetSource(pItem);
		else if(filename)
			hr = link.SetSource(filename);
		else
			return FALSE; // ACepXw肳Ȃ΃G[
		//
		if SUCCEEDED(hr)
		{	// V[gJbĝꍇ͂قƂǖ߁AIɃN擾B
			if(FAILED(link.GetDestination((ITEMIDLIST**)&pItemBuffer.m_p))
			|| FAILED(link.GetDestination(pathBuffer, MAX_PATH)))
			{	// ʂ͂肦ȂB
				MessageBox(filename, _T("V[gJbgt@CjĂ܂B"));
				return FALSE;
			}
			pItem = pItemBuffer;
			filename = pathBuffer;
		}
		//
		if(!pItem && PathFileExists(filename))
		{	// ACeID̂Ŏ擾
			pItem = pItemBuffer = CShellItemTask::FromPath(filename);
		}
		if(!filename && pItem && pItem->GetPath(pathBuffer))
		{	// 
			filename = pathBuffer;
		}

		// t@CGg쐬
		FileNode node;
		ZeroMemory(&node, sizeof(FileNode));
		node.pidl = pItem;
		node.path = filename;
		if(pItem)
			pItem->GetDescription(&node.desc);

		// Jt@Ngt@Nǧ
		CFactoryWindow targetDocument;
		UINT nDocs = CountElements(WINDOW_FACTORY);
		INT suitLevel = SUIT_NONE;
		if(document >= nDocs)
		{	// s̃t@Ng^[QbgɂuJv
			// TODO: ̓x̑Ήt@Ngꍇ̏B
			for(UINT i = 0; i < nDocs; i++)
			{
				CFactoryWindow fac = GetElement(WINDOW_FACTORY, i);
				INT level = fac.QuerySuitable(&node);
				if(level > suitLevel)
				{
					suitLevel = level;
					document = i;
					targetDocument = fac;
				}
			}
			if(!targetDocument)
				return FALSE;
		}
		else
		{	// ̃t@Ng^[QbgɂuJv
			targetDocument = (HWND)GetElement(WINDOW_FACTORY, document);
			suitLevel = targetDocument.QuerySuitable(&node);
		}
		// 
		if(suitLevel < max(SUIT_LOWEST, minSuitLevel))
			return FALSE;
		BOOL ok = targetDocument.Open(&node, pane);
		if(ok && node.path)
			AddHistory(node.path);
		return ok;
	}

	void LoadPlugIn()
	{
		WIN32_FIND_DATA find;
		HANDLE hFind = FindFirstFile(PLUGIN_WILDCARD, &find);
		for(BOOL result = (hFind != NULL); result; result = FindNextFile(hFind, &find))
		{
			TCHAR path[MAX_PATH];
			lstrcpy(path, PLUGIN_DIRECTORY);
			PathAppend(path, find.cFileName);
			HINSTANCE hPlugIn = ::LoadLibrary(path);
			if(!hPlugIn)
				continue;
			TabletInitialize_t fn = (TabletInitialize_t)::GetProcAddress(hPlugIn, EXPORTED_FUNCTION_INITIALIZE);
			if(!fn || !fn(this))
			{
				::FreeLibrary(hPlugIn);
				continue;
			}
			// OK
			m_Libs.Add(hPlugIn);
		}
	}

	void UnloadPlugIn()
	{
		for(UINT i =  0; i < m_Libs.GetCount(); i++)
		{
			TabletTerminate_t fn = (TabletTerminate_t)::GetProcAddress(m_Libs[i], EXPORTED_FUNCTION_TERMINATE);
			if(fn) fn();
			::FreeLibrary(m_Libs[i]);
		}
		m_Libs.RemoveAll();
	}

	////////////////////////////////////////////////////////////////////////////////
	// Operations
public:
	void ParseCommandLine(LPCTSTR lpszCmdLine)
	{
		if(lpszCmdLine == NULL)
			return;
		int len = lstrlen(lpszCmdLine);
		if(len == 0)
			return;
		MessageBox(lpszCmdLine, _T("R}hC"));
	}

	void UIUpdateAll()
	{
		UISetCheck(ID_VIEW_MENUBAR		, IsToolEnabled(WINDOW_MENUBAR));
		UISetCheck(ID_VIEW_TOOLBAR		, IsToolEnabled(WINDOW_TOOLBAR));
		UISetCheck(ID_VIEW_ADDRESSBAR	, IsToolEnabled(WINDOW_ADDRESSBAR));
		UISetCheck(ID_VIEW_STATUS_BAR	, IsToolEnabled(WINDOW_STATUSBAR));
		UISetCheck(ID_VIEW_SIDEBAR		, IsToolEnabled(WINDOW_SIDEBAR));
	}

public: // bZ[W}bv
	BEGIN_UPDATE_UI_MAP(CMainFrame)
		UPDATE_ELEMENT(ID_VIEW_MENUBAR		, UPDUI_MENUPOPUP)
		UPDATE_ELEMENT(ID_VIEW_TOOLBAR		, UPDUI_MENUPOPUP)
		UPDATE_ELEMENT(ID_VIEW_ADDRESSBAR	, UPDUI_MENUPOPUP)
		UPDATE_ELEMENT(ID_VIEW_STATUS_BAR	, UPDUI_MENUPOPUP)
		UPDATE_ELEMENT(ID_VIEW_SIDEBAR		, UPDUI_MENUPOPUP)
		UPDATE_ELEMENT(ID_PANE_MAXIMIZE		, UPDUI_MENUPOPUP | UPDUI_TOOLBAR)
		UPDATE_ELEMENT(ID_PANE_NEW			, UPDUI_MENUPOPUP | UPDUI_TOOLBAR)
	END_UPDATE_UI_MAP()

	BEGIN_MSG_MAP(CMainFrame)
		MESSAGE_HANDLER(WM_CREATE				, OnCreate)
		MESSAGE_HANDLER(WM_DESTROY				, OnDestroy)
		MESSAGE_HANDLER(WM_CLOSE				, OnClose)
		MESSAGE_HANDLER(WM_SETFOCUS				, OnSetFocus)
		MESSAGE_HANDLER(WM_SETTINGCHANGE		, OnSettingChange)
		MESSAGE_HANDLER(WM_TABLET_ELEMENTCHANGE	, OnElementChange)
		MESSAGE_HANDLER(WM_SYSCOMMAND			, OnSysCommand)
		COMMAND_ID_LAMBDA(ID_APP_ABOUT			, { CAboutDlg().DoModal(); })
		COMMAND_ID_LAMBDA(ID_APP_EXIT			, { PostMessage(WM_CLOSE); })
		COMMAND_ID_LAMBDA(ID_FILE_NEW			, { m_NewDlg.DoModal(m_hWnd);  })
		COMMAND_ID_LAMBDA(ID_FILE_OPEN			, { m_OpenDlg.DoModal(m_hWnd); })
		COMMAND_ID_LAMBDA(ID_NEXT_PANE			, { OnPaneNext(true); })
		COMMAND_ID_LAMBDA(ID_PREV_PANE			, { OnPaneNext(false); })
		COMMAND_ID_LAMBDA(ID_PANE_TILE			, { m_PaneSplitter.TilePanes();})
		COMMAND_ID_LAMBDA(ID_PANE_NEW			, { m_PaneSplitter.AddPane(); })
		COMMAND_ID_LAMBDA(ID_VIEW_MENUBAR		, { ToggleToolEnabled(WINDOW_MENUBAR); })
		COMMAND_ID_LAMBDA(ID_VIEW_TOOLBAR		, { ToggleToolEnabled(WINDOW_TOOLBAR); })
		COMMAND_ID_LAMBDA(ID_VIEW_ADDRESSBAR	, { ToggleToolEnabled(WINDOW_ADDRESSBAR); })
		COMMAND_ID_LAMBDA(ID_VIEW_STATUS_BAR	, { ToggleToolEnabled(WINDOW_STATUSBAR); })
		COMMAND_ID_LAMBDA(ID_VIEW_SIDEBAR		, { ToggleToolEnabled(WINDOW_SIDEBAR); })
		COMMAND_ID_LAMBDA(ID_VIEW_SHOW_BUTTON	, { m_ToolBar.OnDropDownShow(); })
		COMMAND_ID_LAMBDA(ID_FOCUS_NEXT			, { OnFocusNext(true); })
		COMMAND_ID_LAMBDA(ID_FOCUS_PREV			, { OnFocusNext(false); })
		COMMAND_ID_HANDLER(ID_PANE_MAXIMIZE		, OnPaneMaximize)
		COMMAND_ID_HANDLER(ID_PANE_DIRECTION	, OnPaneDirection)
		COMMAND_ID_HANDLER(ID_TOOL_OPTION		, OnToolOption)
		NOTIFY_HANDLER(IDW_ADDRESSBAR, CBEN_ENDEDIT	, OnAddressEndEdit)
		// must be MenuProvider => UpdateUI => MenuHandler => ... => OnCommand
		CHAIN_MSG_MAP(baseMenuProvider)
		CHAIN_MSG_MAP(baseUpdateUI)
		CHAIN_MSG_MAP(baseMenuHandler)
		CHAIN_MSG_MAP(baseClass)
		MESSAGE_HANDLER(WM_COMMAND	, OnCommand)
		REFLECT_NOTIFICATIONS()
	ALT_MSG_MAP(MSG_MAP_ID_GESTURE)
		CHAIN_MSG_MAP_ALT(baseGesture, MSG_MAP_ID_GESTURE)
	END_MSG_MAP()

//	LRESULT MessageHandler(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/)
//	LRESULT CommandHandler(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
//	LRESULT NotifyHandler(int /*idCtrl*/, LPNMHDR /*pnmh*/, BOOL& /*bHandled*/)

	// fXgN^ł͂ȂAEBhE̔jɃCuB
	void OnFinalMessage(HWND)
	{
		UnloadPlugIn();
		PostQuitMessage(0); // WM_DESTORY ł͂AÔ߂x
	}

	BOOL PreTranslateMessage(MSG* pMsg)
	{
		if(baseWheel::PreTranslateMessage(pMsg))
			return TRUE;

		CPaneWindow pane = m_PaneSplitter.GetActiveElement();
		CWindow view = pane.GetActiveElement();
		if(pMsg->message == WM_RBUTTONDOWN && view && (view == pMsg->hwnd || view.IsChild(pMsg->hwnd)))
		{
			CPoint pt(pMsg->lParam);
			::ClientToScreen(pMsg->hwnd, &pt);
			if((BOOL)view.SendMessage(WM_TABLET_BEGINGESTURE, pMsg->wParam, MAKELPARAM(pt.x, pt.y)))
			{
				BeginGesture(pMsg);
				return TRUE;
			}
		}

		if(pane.PreTranslateMessage(pMsg))
			return TRUE;
		if(m_SidePane.PreTranslateMessage(pMsg))
			return TRUE;
		return baseClass::PreTranslateMessage(pMsg);
	}

	BOOL OnIdle()
	{
		UIUpdateToolBar();
		UIEnable(ID_PANE_NEW, CountElements(WINDOW_PANESPLITTER) < MAX_PANE_COUNT);
		UISetCheck(ID_PANE_MAXIMIZE, m_PaneSplitter.IsSinglePaneMode());
		return FALSE;
	}

	LRESULT OnCreate(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/)
	{
		// Documents
		m_FactoryPane.Create(m_hWnd, rcDefault, _T("WINDOW_FACTORY"), 0, 0, WINDOW_FACTORY);

		// ReBar/CommandBar/ToolBar/AddressBar/StatusBar
		m_hWndToolBar = m_ReBar.Create(m_hWnd, IDW_REBAR);
		m_MenuBar.Create(m_hWnd, IDW_MENUBAR);
		m_ToolBar.Create(m_hWnd, IDW_TOOLBAR);
		m_AddressBar.Create(m_hWnd, IDW_ADDRESSBAR, Default::AddressBarHeight);
		m_hWndStatusBar = m_StatusBar.Create(m_hWnd, ATL_IDW_STATUS_BAR);

/*
		// TabBar
		CImageList images;
		images.Create(IDR_MAINFRAME, 16, 10, 0);
		m_TabBar.Create(m_hWnd, CRect(0, 0, 0, 40), NULL, WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS | TCS_BUTTONS | TCS_MULTISELECT | TCS_FOCUSNEVER | TCS_FLATBUTTONS | TCS_HOTTRACK );
//		m_TabBar.SetExtendedStyle(0, TCS_EX_REGISTERDROP);
		m_TabBar.SetImageList(images);
		TCITEM item = { TCIF_TEXT | TCIF_IMAGE };
		item.pszText = _T(" 1");
		item.iImage = 0;
		m_TabBar.InsertItem(0, &item);
		item.pszText = _T(" 2");
		item.iImage = 1;
		m_TabBar.InsertItem(0, &item);
		item.pszText = _T(" 3");
		item.iImage = 2;
		m_TabBar.InsertItem(0, &item);
		AddSimpleReBarBand(m_TabBar, _T("^u"), TRUE);
*/
		// t[̃NCAg̓Xvb^B
		m_hWndClient = m_RootSplitter.Create(m_hWnd);
		m_RootSplitter.SetSplitterExtendedStyle(0); // 0 means SPLIT_LEFTALIGNED
		// Xvb^̎qij̓TChyCB
		m_SidePane.Create(m_RootSplitter, rcDefault, _T("WINDOW_SIDEBAR"), 0, 0, WINDOW_SIDEBAR);
		// Xvb^̎qiEj̓_Ci~bNXvb^B
		m_PaneSplitter.Create(m_RootSplitter, rcDefault, _T("WINDOW_PANESPLITTER"), 0, 0, WINDOW_PANESPLITTER);
		// yCPǉB
		m_PaneSplitter.EnsurePaneCount(1);
		// Xvb^ɓo^B
		m_RootSplitter.SetSplitterPanes(m_SidePane, m_PaneSplitter);

		// SysMenu
		CMenuHandle sysmenu = GetSystemMenu(FALSE);
		sysmenu.InsertMenu(0, MF_ENABLED, ID_VIEW_MENUBAR, _T("j[\/\"));

		// Drag&Drop
		RegisterDragDrop();

		// register object for message filtering and idle updates
		CMessageLoop* pLoop = _Module.GetMessageLoop();
		ATLASSERT(pLoop != NULL);
		pLoop->AddMessageFilter(this);
		pLoop->AddIdleHandler(this);

		// Plug-in
		LoadPlugIn();

		// Profile
		CProfile profile(theProfile, _T("MainFrame"));
		LoadProfile(profile);

		m_StatusBar.SetPaneWidth(ID_SECOND_PANE, 125);
		m_MenuBar.PostCreate(m_ToolBar);

		// UI
		UIAddToolBar(m_ToolBar);
		UIUpdateAll();

		return 0;
	}

	static BOOL CALLBACK DestroyChildrenProc(HWND hwnd, LPARAM lParam)
	{
		::DestroyWindow(hwnd);
		return TRUE;
	}

	LRESULT OnDestroy(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled)
	{
		// Profile
		CProfile profile(theProfile, _T("MainFrame"));
		SaveProfile(profile);
		//
		::EnumChildWindows(m_hWnd, DestroyChildrenProc, 0);
		// ȂƁAVbg_EɃbZ[WzMĂ܂B
		CMessageLoop* pLoop = _Module.GetMessageLoop();
		ATLASSERT(pLoop != NULL);
		pLoop->RemoveMessageFilter(this);
		pLoop->RemoveIdleHandler(this);
		bHandled = FALSE;
		return 0;
	}

	LRESULT OnClose(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled)
	{
		// TODO: ꂼ̃r[ WM_CLOSE 𑗂
		//if(ׂẴr[ӂ)
		bHandled = FALSE;
		return 0;
	}

	LRESULT OnSetFocus(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/)
	{
		CPaneWindow pane = m_PaneSplitter.GetActiveElement();
		if(pane.CountElements() > 0)
			pane.SetFocus();
		else if(m_SidePane.CountElements() > 0)
			m_SidePane.SetFocus();
		else
			m_AddressBar.SetFocus();
		return 0;
	}

	LRESULT OnSettingChange(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& /*bHandled*/)
	{
		SendMessageToDescendants(uMsg, wParam, lParam, TRUE);
		return 0;
	}

	LRESULT OnElementChange(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& /*bHandled*/)
	{
		SendMessageToDescendants(uMsg, wParam, lParam, TRUE);
		HWND hWnd = (HWND)lParam;
#if 0 //def _DEBUG
		const LPCTSTR MSGSTR[] =
		{
			_T("CHANGE_FACTORY_ADD"),		// t@Ngǉ܂
			_T("CHANGE_FACTORY_DELETE"),	// t@Ng폜܂
			_T("CHANGE_FACTORY_MOVE"),		// t@Ng̕яω܂
			_T("CHANGE_SIDEBAR_ADD"),		// TCho[ǉ܂
			_T("CHANGE_SIDEBAR_MOVE"),		// TCho[폜܂
			_T("CHANGE_SIDEBAR_DELETE"),	// TCho[̕яω܂
			_T("CHANGE_SIDEBAR_ACTIVE"),	// ANeBuTCho[ω܂
			_T("CHANGE_PANE_ADD"),			// yCǉ܂
			_T("CHANGE_PANE_MOVE"),			// yC폜܂
			_T("CHANGE_PANE_DELETE"),		// yC̕яω܂
			_T("CHANGE_PANE_ACTIVE"),		// ANeBuyCω܂
			_T("CHANGE_VIEW_ADD"),			// r[ǉ܂
			_T("CHANGE_VIEW_MOVE"),			// r[폜܂
			_T("CHANGE_VIEW_DELETE"),		// r[̕яω܂
			_T("CHANGE_VIEW_ACTIVE"),		// ANeBur[ω܂
		};
		TCHAR name[MAX_PATH] = _T("*ERROR*");
		if(::IsWindow(hWnd))
		{
			if(!::SendMessage(hWnd, WM_TABLET_GETNAMES, 0, (LPARAM)name))
				::GetWindowText(hWnd, name, MAX_PATH);
		}
		ATLTRACE(_T("%s : %s [0x%08X]\n"), MSGSTR[wParam], name, lParam);
#endif
		switch(wParam)
		{
		case CHANGE_VIEW_ADD:
		case CHANGE_VIEW_MOVE:
			if(true /* Vr[ǉꂽꍇɃANeBuɂ */)
				m_PaneSplitter.SetActivePaneWindow(hWnd);
			break;
		case CHANGE_VIEW_DELETE:
			break;
		case CHANGE_PANE_ACTIVE:
			// CHANGE_PANE_ACTIVE ɑ΂āAANeBuɂȂyC
			// CHANGE_VIEW_ACTIVE đقH
			OnUpdateAddress(CPaneWindow(hWnd).GetActiveElement());
			break;
		case CHANGE_VIEW_ACTIVE:
			OnUpdateAddress(hWnd);
			break;
		}
		return 0;
	}

	void OnUpdateAddress(CCommonWindow window)
	{
		TCHAR shortName[MAX_PATH] = _T(""), longName[MAX_PATH] = _T("");
		if(window)
			window.GetNames(shortName, longName);
		// AhXo[XV
		m_AddressBar.Set(longName);
		// ^CgXV
		int len = lstrlen(shortName);
		CString title;
		title.LoadString(IDR_MAINFRAME);
		if(len > 0)
		{
			CString caption;
			caption.Preallocate(len + 3 + title.GetLength()); // "ShortName - Title"
			caption.LoadString(IDR_MAINFRAME);
			caption = shortName;
			caption.Append(_T(" - "));
			caption.Append(title);
			this->SetWindowText(caption);
		}
		else
			this->SetWindowText(title);
	}

	LRESULT OnCommand(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& /*bHandled*/)
	{
		// Ct[ȂR}hANeBuȃr[yCɓnB
		// ɁAr[yC̓ANeBuȃr[ɓnB
		return m_PaneSplitter.GetActiveElement().SendMessage(uMsg, wParam, lParam);
	}

	LRESULT OnSysCommand(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		// VXej[̏ꍇ wParam & 0xFFF0 Ŕr邪A[U[j[̏ꍇ͐̒lŔrB
		switch(wParam)
		{
		case ID_VIEW_MENUBAR:
			SendMessage(WM_COMMAND, MAKEWPARAM(ID_VIEW_MENUBAR, 0), 0);
			break;
		default:
			bHandled = FALSE;
			break;
		}
		return 0;
	}

	void OnPaneNext(bool next)
	{
		if(m_PaneSplitter.IsSinglePaneMode() || m_PaneSplitter.GetPaneCount() <= 1)
		{	// VO or yCꖇꍇ
			CPaneWindow pane = m_PaneSplitter.GetActiveElement();
			INT index = pane.GetActiveIndex();
			if(index >= 0)
				pane.SetActiveIndex(index + (next ? 1 : -1));
		}
		else
			m_PaneSplitter.ActivateNextPane(next);
	}

	LRESULT OnPaneMaximize(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
	{
		if(m_PaneSplitter.IsSinglePaneMode())
			m_PaneSplitter.SetSinglePaneMode(SPLIT_PANE_NONE);
		else
			m_PaneSplitter.SetSinglePaneMode(m_PaneSplitter.GetActivePane());
		return 0;
	}

	LRESULT OnPaneDirection(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
	{
		m_PaneSplitter.SetVertical(!m_PaneSplitter.IsVertical());
		m_PaneSplitter.TilePanes();
		return 0;
	}

	LRESULT OnToolOption(WORD /*wNotifyCode*/, WORD wID, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
	{
		COptionProp option;
		option.DoModal(m_hWnd);
		return 0;
	}

	LRESULT OnAddressEndEdit(int /*idCtrl*/, LPNMHDR pnmh, BOOL& /*bHandled*/)
	{
		NMCBEENDEDIT* edit = (NMCBEENDEDIT*)pnmh;
		switch(edit->iWhy)
		{
		case CBENF_RETURN:
			if(lstrlen(edit->szText) > 0)
			{
				if(theTablet->OpenFile(edit->szText, WINDOW_PANE_ACTIVE, -1))
				{
					// OK
				}
				else
					MessageBox(edit->szText, _T("̃hLgJvOC܂"));
			}
			return FALSE;
		default:
			return TRUE;
		}
	}

	void OnFocusNext(bool next)
	{
		HWND hNextFocus[] = { m_AddressBar.m_hWnd, m_SidePane.m_hWnd, m_PaneSplitter.m_hWnd, m_AddressBar.m_hWnd, m_SidePane.m_hWnd };
		HWND hFocus = ::GetFocus();
		for(int i = 1; i < 4; i++)
		{
			if(hNextFocus[i] == hFocus || ::IsChild(hNextFocus[i], hFocus))
			{
				::SetFocus(hNextFocus[i + (next ? 1 : -1)]);
				return;
			}
		}
		// tH[JXȂ̂ŁAtH[JXɔCB
		this->SetFocus();
	}

public: // CDropTargetImpl
	BEGIN_FORMAT_MAP(CMainFrame)
		FORMAT_STANDARD_FILES()
	END_FORMAT_MAP()

	void OnDrop(IDataObject* pDataObject, DWORD key, CPoint ptScreen, DWORD& effect)
	{
		ProcessDropFiles(pDataObject, key, ptScreen, effect);
	}
	void OnDropFile(LPCTSTR filename, const CShellItem* pItem, DWORD key, CPoint ptScreen, DWORD& effect)
	{
		CPoint pt = ptScreen;
		m_PaneSplitter.ScreenToClient(&pt);
		INT index = m_PaneSplitter.HitTest(pt);
		// qbgeXgs͂ -1 Ԃ̂ŁA傤 WINDOW_PANE_ACTIVE ɂȂB
		if(filename)
			OpenFile(filename, (WindowID)(WINDOW_PANE_FIRST+index), -1, SUIT_NONE);
		else if(pItem)
			OpenFile(pItem, (WindowID)(WINDOW_PANE_FIRST+index), -1, SUIT_NONE);
	}

public: // CMouseGestureImpl
	void OnGestureBegin()
	{
		m_StatusBar.SetSimple(TRUE);
	}

	void OnGestureEnd()
	{
		m_StatusBar.SetSimple(FALSE);
	}

	void OnGestureUpdate(UINT length, const Gesture* gesture)
	{
		CString str;
		str.Preallocate(256);
		str = _T("}EXWFX` : ");
		for(UINT i = 0; i < length; i++)
		{
			str += GESTURE_CHAR[gesture[i]];
		}
		m_StatusBar.SetText(255 , str, SBT_NOBORDERS);
	}

	void OnGestureFinish(UINT length, const Gesture* gesture)
	{
		CWindow active = GetWindow(WINDOW_VIEW_ACTIVE);
		if(active)
		{
			active.SendMessage(WM_TABLET_ENDGESTURE, (WPARAM)length, (LPARAM)gesture);
		}
	}
};
