// ShellUtil.h
#pragma once

#include "Allocator.h"
#include "DataObject.h"
#include "Collection.h"

namespace mew {

/////////////////////////////////////////////////////////////////////////////
// Classes in this file
//
// CShellLink			: IShellLink wrapper
// CShellItem			: ITEMIDLIST wrapper
// CShellItemT			: CShellItem with memory allocation
// CShellItemList		: CIDA wrapper
// CShellItemListT		: CShellItemList with memory allocation
// CDataObject_ShellItemList

/////////////////////////////////////////////////////////////////////////////
// typedefs
// 
// CShellItemTask	= CShellItemT<CTaskAllocator>
// CShellItemListGlobal	= CShellItemListT<CGlobalLockAllocator>

////////////////////////////////////////////////////////////////////////

enum FileEnumFlag
{
	FileEnumFlag_IncludeFolders		= 0x0020,	// == SHCONTF_FOLDERS
	FileEnumFlag_IncludeFiles		= 0x0040,	// == SHCONTF_NONFOLDERS
	FileEnumFlag_IncludeHidden		= 0x0080,	// == SHCONTF_INCLUDEHIDDEN
	FileEnumFlag_SortFoldersOnTop	= 0x1000,
	FileEnumFlag_MaskSHCONTF		= 0x0FF0,
};

////////////////////////////////////////////////////////////////////////

class CShellLink
{
private:
	CComPtr<IShellLink> m_pShellLink;

public:
	HRESULT SetSource(LPCSTR filename)
	{
		return SetSource(CA2W(filename));
	}
	HRESULT SetSource(LPCWSTR filename)
	{
		m_pShellLink.Release();
		CComPtr<IPersistFile> pPersistFile;
		HRESULT hr;
		if(FAILED(hr = m_pShellLink.CoCreateInstance(CLSID_ShellLink))
		|| FAILED(hr = m_pShellLink.QueryInterface(&pPersistFile))
		|| FAILED(hr = pPersistFile->Load(filename, STGM_READ)))
		{
			return hr;
		}
		else
			return S_OK;
	}
	HRESULT SetSource(LPCITEMIDLIST pIDL)
	{
		m_pShellLink.Release();
		TCHAR path[MAX_PATH];
		if(!SHGetPathFromIDList(pIDL, path))
			return E_FAIL;
		else
			return SetSource(path);
	}
	HRESULT GetDestination(LPTSTR path, int cch)
	{
		if(!m_pShellLink)
			return E_FAIL;
		return m_pShellLink->GetPath(path, cch, 0, 0);
	}
	HRESULT GetDestination(LPITEMIDLIST* ppIDL)
	{
		if(!m_pShellLink)
			return E_FAIL;
		return m_pShellLink->GetIDList(ppIDL);
	}
};

////////////////////////////////////////////////////////////////////////

class CShellItem : public ITEMIDLIST
{
private:
	CShellItem();
	CShellItem(const CShellItem&);
	const CShellItem& operator = (const CShellItem&);
protected:
	~CShellItem()	{} // forbid operator delete

public:
	static const CShellItem* GetNextElement(LPCITEMIDLIST pIDL)
	{
		return pIDL ? (const CShellItem*)(BYTE*)(((BYTE*)pIDL) + pIDL->mkid.cb) : NULL;
	}
	const CShellItem* GetNextElement() const	{ return GetNextElement(this); }
	CShellItem* GetNextElement()				{ return const_cast<CShellItem*>(GetNextElement(this)); }

	// [m[hԂB̃m[hthisƓobt@wĂ̂ŁAĂ͂ȂȂB
	// pParentSize, pLeafSize ́A^[~l[^܂񂾃TCYԂB
	// ̂߁Athis->GetSize() + 2 == *pParentSize + *pLeafSize ɂȂB
	CShellItem* FindLeaf(UINT* pParentSize = NULL, UINT* pLeafSize = NULL)
	{
		UINT parentSize = 2, leafSize = 2;// [ (WORD)0 ŋ؂邽߁B
		UINT sz = 0;
		CShellItem* leaf = NULL;
		for(CShellItem* pidl = this; pidl->mkid.cb; pidl = pidl->GetNextElement())
		{
			sz = pidl->mkid.cb;
			parentSize += sz;
			leaf = pidl;
		}
		if(leaf == NULL)
		{	// etH_Ɩ[ɕłȂfXNgbvȁH
			leafSize = 0;
			return NULL;
		}
		parentSize -= sz;
		leafSize += sz;
		if(pParentSize) *pParentSize = parentSize;
		if(pLeafSize)   *pLeafSize   = leafSize;
		return leaf;
	}

	BOOL GetPath(LPTSTR path) const
	{
		return SHGetPathFromIDList(this, path);
	}

	const CShellItem* FindLeaf(UINT* pParentSize = NULL, UINT* pLeafSize = NULL) const
	{
		return const_cast<CShellItem*>(this)->FindLeaf(pParentSize, pLeafSize);
	}

	static UINT GetSize(LPCITEMIDLIST pIDL)
	{
		if(!pIDL)
			return 0;
		// Requires a 16 bit zero value for the NULL terminator
		UINT cbTotal = 2;
		for(LPCITEMIDLIST pidl = pIDL; pidl->mkid.cb; pidl = GetNextElement(pidl))
		{
			cbTotal += pidl->mkid.cb;
		}  
		return cbTotal;
	}

	UINT GetSize() const
	{
		return GetSize(this);
	}

	static bool Equals(LPCITEMIDLIST lhs, LPCITEMIDLIST rhs)
	{
		UINT szL = GetSize(lhs);
		UINT szR = GetSize(rhs);
		if(szL != szR)
			return false;
		return memcmp(lhs, rhs, szL) == 0;
	}

	bool Equals(LPCITEMIDLIST rhs) const
	{
		return Equals(this, rhs);
	}

	bool IsRoot() const
	{
		// WindowsXPł́AfXNgbv͋̃m[hA
		// ɂĂ͈قȂ邩ȂH
		return mkid.cb == 0;
	}

	HRESULT RemoveLeaf(CShellItem* pLeafHint = NULL)
	{
		if(!pLeafHint && !(pLeafHint = FindLeaf()))
		{
			return E_FAIL;
		}
		pLeafHint->mkid.cb = 0;
		return S_OK;
	}

	HRESULT GetParentFolder(IShellFolder** ppParent, CShellItem** ppRelative = NULL) const
	{
		// ppRelative; // ԂA̒lpFullPatḧꕔwĂ遁Ȃėǂ
		return SHBindToParent(this, IID_IShellFolder, (void**)ppParent, (LPCITEMIDLIST*)ppRelative);
	}

	HRESULT GetSelfFolder(IShellFolder** ppSelfFolder) const
	{
		HRESULT hr;
		if(IsRoot())
			return SHGetDesktopFolder(ppSelfFolder);
		CShellItem* pRelative;
		CComPtr<IShellFolder> pParent;
		if(SUCCEEDED(hr = GetParentFolder(&pParent, &pRelative))
		&& SUCCEEDED(hr = pParent->BindToObject(pRelative, NULL, IID_IShellFolder, (void**)ppSelfFolder)))
			return S_OK;
		else
			return hr;
	}

	HRESULT GetSelfFolderRelative(IShellFolder** ppSelfFolder, IShellFolder* pParent) const
	{
		if(SUCCEEDED(pParent->BindToObject(this, NULL, IID_IShellFolder, (void**)ppSelfFolder)))
			return S_OK;
		else
			return SHGetDesktopFolder(ppSelfFolder);
	}

	HRESULT GetDescription(SHDESCRIPTIONID* desc) const
	{
		CShellItem* pRelative;
		CComPtr<IShellFolder> pParent;
		GetParentFolder(&pParent, &pRelative);
		return SHGetDataFromIDList(pParent, pRelative, SHGDFIL_DESCRIPTIONID, desc, sizeof(SHDESCRIPTIONID));
	}

	DWORD_PTR GetFileInfo(SHFILEINFO* file, DWORD flags) const
	{
		return ::SHGetFileInfo((LPCTSTR)this, 0, file, sizeof(SHFILEINFO), flags | SHGFI_PIDL);
	}

	HRESULT GetProperty(LPTSTR name, DWORD* attr = NULL, CShellItem** ppLeaf = NULL) const
	{
		HRESULT hr;
		CComPtr<IShellFolder> pParent;
		CShellItem* pLeaf;
		if(FAILED(hr = GetParentFolder(&pParent, &pLeaf)))
			return hr;
		if(ppLeaf) *ppLeaf = pLeaf;
		return pLeaf->GetPropertyOfLeaf(pParent, name, attr);
	}

	HRESULT GetPropertyOfLeaf(IShellFolder* pParentFolder, LPTSTR name, DWORD* attr) const
	{
		if(attr)
		{
			LPCITEMIDLIST pidl = this;
			pParentFolder->GetAttributesOf(1, &pidl, attr);
		}
		if(name)
		{
			STRRET strret;
			pParentFolder->GetDisplayNameOf(this, SHGDN_NORMAL, &strret);
			StrRetToBuf(&strret, this, name, MAX_PATH);
		}
		return S_OK;
	}

	HRESULT GetUIObjectOf(HWND hWnd, REFIID riid, void** ppObject) const
	{
		CShellItem* pLeaf;
		CComPtr<IShellFolder> pParentFolder;
		if FAILED(GetParentFolder(&pParentFolder, &pLeaf))
		{
			SHGetDesktopFolder(&pParentFolder);
			LPCITEMIDLIST pThis = this;
			return pParentFolder->GetUIObjectOf(hWnd, 1, &pThis, riid, NULL, ppObject);
		}
		else
		{
			return pParentFolder->GetUIObjectOf(hWnd, 1, (LPCITEMIDLIST*)&pLeaf, riid, NULL, ppObject);
		}
	}

	template < class U >
	HRESULT GetUIObjectOf(HWND hWnd, U** ppObject, REFIID riid = __uuidof(U)) const
	{
		return GetUIObjectOf(hWnd, riid, (void**)ppObject);

	}

	BOOL ShellExecute(HWND hWnd, LPCTSTR directory = NULL) const
	{
		SHELLEXECUTEINFO info = { sizeof(SHELLEXECUTEINFO), SEE_MASK_IDLIST, hWnd };
		info.nShow = SW_SHOWNORMAL;
		info.lpIDList = (LPITEMIDLIST)this;
		TCHAR dir[MAX_PATH];
		if(directory == NULL && this->GetPath(dir))
		{	// [LOfBNǵANt@CƓ
			PathRemoveFileSpec(dir);
			info.lpDirectory = dir;
		}
		return ShellExecuteEx(&info);
	}
};

////////////////////////////////////////////////////////////////////////

template < class TAlloc >
class CShellItemT : public CShellItem
{
private:
	CShellItemT();
	CShellItemT(const CShellItemT&);
	const CShellItemT& operator = (const CShellItemT&);

public:
	void operator delete(void* ptr)
	{
		TAlloc::Free(ptr);
	}

	static CAutoPtr<CShellItemT> FromPath(LPCTSTR path)
	{
		CAutoPtr<CShellItemT> result;
		// ܂̓fXNgbv擾.
		CComPtr<IShellFolder> pDesktopFolder; 
		if FAILED(::SHGetDesktopFolder(&pDesktopFolder))
            return result;
		// 
		ITEMIDLIST* pIDL;
		ULONG chEaten;	//̃TCY󂯎܂B
		ULONG dwAttributes;	//󂯎܂B
		if FAILED(pDesktopFolder->ParseDisplayName(NULL, NULL, CT2OLE(path), &chEaten, &pIDL, &dwAttributes))
			return result;
		result.Attach((CShellItemT*)pIDL);
		return result;
	}

	static CAutoPtr<CShellItemT> FromCSIDL(INT id, HWND hWndOwner)
	{
		CAutoPtr<CShellItemT> ret;
		ITEMIDLIST* pIDL;
		HRESULT hr = SHGetFolderLocation(hWndOwner, id, 0, 0, &pIDL);
		if FAILED(hr)
		{
			ATLTRACE(_T("ERROR: SHGetFolderLocation(%d)\n"), id);
			return ret;
		}
		ret.Attach((CShellItemT*)pIDL);
		return ret;
	}

	static CAutoPtr<CShellItemT> Concat(LPCITEMIDLIST pidl1, LPCITEMIDLIST pidl2)
	{
		UINT cb1 = 0, cb2 = 0;
		if(pidl1)
			cb1 = GetSize(pidl1) - 2;
		cb2 = GetSize(pidl2);
		CShellItemT* pidlNew = (CShellItemT*)TAlloc::Allocate(cb1 + cb2);
		ZeroMemory(pidlNew, cb1 + cb2);
		if(pidl1)
			CopyMemory(pidlNew, pidl1, cb1);
		CopyMemory(((BYTE*)pidlNew) + cb1, pidl2, cb2);
		return CAutoPtr<CShellItemT>(pidlNew);
	}

	static CAutoPtr<CShellItemT> Duplicate(LPCITEMIDLIST pIDL)
	{
		UINT size = GetSize(pIDL);
		CAutoPtr<CShellItemT> list((CShellItemT*)TAlloc::Allocate(size));
		CopyMemory(list, pIDL, size);
		return list;
	}

	CAutoPtr<CShellItemT> GetParent() const
	{
		UINT parentSize = 0;
		FindLeaf(&parentSize);
		CAutoPtr<CShellItemT> pParent((CShellItemT*)TAlloc::Allocate(parentSize));
		ZeroMemory(pParent.m_p, parentSize);
		CopyMemory(pParent.m_p, this, parentSize-2);
		return pParent;
	}

	////////////////////////////////////////////////////////////////////////

	struct EnumData
	{
		CAutoPtr<CShellItemT> pLeaf;
		TCHAR	Name[MAX_PATH];
		DWORD	Attr;

		BOOL IsFolder() const
		{
			return Attr & SFGAO_FOLDER;
		}
	};

	struct Sorter
	{
	private:
		BOOL m_FolderOnTop;
	public:
		Sorter(BOOL folderOnTop) : m_FolderOnTop(folderOnTop)
		{
		}
		bool operator ()(const EnumData* lhs, const EnumData* rhs)
		{
			// FIXME: hCu{[xŃ\[gĂ܂BiC, D, E...ɂȂȂj
			if(m_FolderOnTop)
			{
				if(lhs->IsFolder() && !rhs->IsFolder())
					return true;
				else if(!lhs->IsFolder() && rhs->IsFolder())
					return false;
			}
			return lstrcmpi(lhs->Name, rhs->Name) < 0;
		}
	};

	typedef CAutoPtrVector<EnumData>	EnumResult;

	static HRESULT EnumChildren(IShellFolder* pParentFolder, DWORD dwEnumFlags, EnumResult& out)
	{
		HRESULT hr;
		CComPtr<IEnumIDList> pEnum;
		DWORD flags = (dwEnumFlags & FileEnumFlag_MaskSHCONTF);
		if FAILED(hr = pParentFolder->EnumObjects(NULL, flags, &pEnum))
			return hr;
		CShellItemT* leaf;
		while(pEnum->Next(1, (LPITEMIDLIST*)&leaf, NULL) == S_OK)
		{
			CAutoPtr<EnumData> item(new EnumData);
			item->pLeaf.Attach(leaf);
			item->Attr  = SFGAO_HASSUBFOLDER | SFGAO_FOLDER | SFGAO_DISPLAYATTRMASK | SFGAO_CANRENAME;
			leaf->GetPropertyOfLeaf(pParentFolder, item->Name, &item->Attr);
			out.push_back(item);
		}
		//
		if(out.empty())
			return S_FALSE;
		// \[g
		std::sort(out.begin(), out.end(), Sorter(dwEnumFlags & FileEnumFlag_SortFoldersOnTop));
		return S_OK;
	}
};

typedef CShellItemT<CTaskAllocator>	CShellItemTask;

////////////////////////////////////////////////////////////////////////

class CShellItemList : public CIDA
{
private:
	CShellItemList();
	CShellItemList(const CShellItemList&);
	const CShellItemList& operator = (const CShellItemList&);
protected:
	~CShellItemList()	{} // forbid operator delete

public:
	UINT GetCount() const
	{
		return cidl;
	}
	const CShellItem* GetParent() const
	{
		return (CShellItem*)(((BYTE*)this)+aoffset[0]);
	}
	const CShellItem* GetAt(UINT i) const
	{
		return (CShellItem*)(((BYTE*)this)+aoffset[i+1]);
	}
	UINT GetSize() const
	{
		return aoffset[cidl] + GetAt(cidl-1)->GetSize();
	}
	bool CopyTo(void* buffer) const
	{
		CopyMemory(buffer, this, GetSize());
		return true;
	}
	HGLOBAL CreateHDROP() const
	{
		// HDROP ́ADROPFILES \̂̃wb_̌ɁA_uNULL؂t@Cz񂪑`B
		// ccƂƂ́ADragQueryFile() ͎w肵CfbNX̃t@C擾̂
		// _uNULL؂蕶XLĂ̂ccB
		CComPtr<IShellFolder> pParentFolder;
		if FAILED(GetParent()->GetSelfFolder(&pParentFolder))
			return NULL;
		TCHAR parentPath[MAX_PATH];
		GetParent()->GetPath(parentPath);
		PathAddBackslash(parentPath);
		UINT parentPathLength = lstrlen(parentPath);
		// ̂Ȃ񂠂Ώ\낤B
		UINT allocateBytes = sizeof(DROPFILES) + (parentPathLength+64) * GetCount() * sizeof(TCHAR);
		HGLOBAL hGlobal = ::GlobalAlloc(GMEM_MOVEABLE, allocateBytes);
		CComPtr<IStream> pStream;
		CreateStreamOnHGlobal(hGlobal, FALSE, &pStream);
		DROPFILES dropfiles =
		{
			sizeof(DROPFILES), 0, 0, TRUE,
#ifdef _UNICODE
			TRUE
#else
			FALSE
#endif
		};
		GetCursorPos(&dropfiles.pt);
		pStream->Write(&dropfiles, sizeof(DROPFILES), NULL);
		for(UINT i = 0; i < GetCount(); i++)
		{
			WIN32_FIND_DATA desc;
			SHGetDataFromIDList(pParentFolder, GetAt(i), SHGDFIL_FINDDATA, &desc, sizeof(desc));
			pStream->Write(parentPath, parentPathLength * sizeof(TCHAR), NULL);
			pStream->Write(desc.cFileName, (lstrlen(desc.cFileName)+1) * sizeof(TCHAR), NULL); // NULL
		}
		pStream->Write(_T("\0"), sizeof(TCHAR), NULL); // Ō̓_uNULL
		GetHGlobalFromStream(pStream, &hGlobal);
		return hGlobal;
	}
};

template < class TAlloc >
class CShellItemListT : public CShellItemList
{
private:
	CShellItemListT();
	CShellItemListT(const CShellItemListT&);
	const CShellItemListT& operator = (const CShellItemListT&);

public:
	void operator delete(void* ptr)
	{
		TAlloc::Free(ptr);
	}

	static CAutoPtr<CShellItemListT> Create(const CShellItem* single)
	{
		CAutoPtr<CShellItemListT> list;
		UINT parentSize, leafSize;
		const CShellItem* leaf = single->FindLeaf(&parentSize, &leafSize);
		if(leaf == NULL)
		{	// etH_Ɩ[ɕłȂfXNgbvȁH
			ATLTRACE("CShellItemListT::Create() failed.\n");
			return list;
		}
		
		ATLASSERT(leaf->GetSize() == leafSize);
		ATLASSERT(single->GetSize()+2 == parentSize + leafSize);

		UINT headerSize = sizeof(UINT)*3;
		UINT totalSize = headerSize + parentSize + leafSize;

		list.Attach((CShellItemListT*)TAlloc::Allocate(totalSize));
		ZeroMemory(list, totalSize);
		list->cidl = 1;
		list->aoffset[0] = headerSize;
		list->aoffset[1] = headerSize + parentSize;
		// [ (WORD)0 Ȃ̂ŁARs[ȂiZeroMemory()łłɃ[߂Ăj
		CopyMemory(((BYTE*)list.m_p) + list->aoffset[0], single, parentSize-2);
		CopyMemory(((BYTE*)list.m_p) + list->aoffset[1], leaf, leafSize-2);

		return list;
	}

	static CAutoPtr<CShellItemListT> Duplicate(const CIDA* pCIDA)
	{
		UINT size = static_cast<const CShellItemList*>(pCIDA)->GetSize();
		CAutoPtr<CShellItemListT> list((CShellItemListT*)TAlloc::Allocate(size));
		CopyMemory(list, pCIDA, size);
		return list;
	}
};

typedef CShellItemListT<CGlobalLockAllocator>	CShellItemListGlobal;

//////////////////////////////////////////////////////////////////////////////

class CDataObject_ShellItemList : public CDataObjectImpl< CDataObject_ShellItemList >
{
private:
	typedef CShellItemListT<CTaskAllocator>	ShellItemList;
	CAutoPtr<ShellItemList> m_pShellItemList;
	typedef CDataObjectImpl< CDataObject_ShellItemList >	BaseClass;

public:
	BEGIN_FORMAT_MAP(CDataObject_ShellItemList)
		FORMAT_ENTRY(CFSTR_SHELLIDLIST, TYMED_HGLOBAL | TYMED_ISTREAM)
		FORMAT_ENTRY(CF_HDROP         , TYMED_HGLOBAL | TYMED_ISTREAM)
	END_FORMAT_MAP()

	CDataObject_ShellItemList(const CShellItem* pSingle)
	{
		m_pShellItemList = ShellItemList::Create(pSingle);
	}
	CDataObject_ShellItemList(const CShellItemList* list)
	{
		m_pShellItemList = ShellItemList::Duplicate(list);
	}

	HRESULT GetDataOnHGlobal(FORMATETC* fmt, HGLOBAL* phGlobal)
	{
		if(!m_pShellItemList)
			return E_FAIL;
		if(fmt->cfFormat == CFormatEtc::ToCLIPFORMAT(CFSTR_SHELLIDLIST))
		{
			size_t size = m_pShellItemList->GetSize();
			*phGlobal = ::GlobalAlloc(GMEM_MOVEABLE, size);
			m_pShellItemList->CopyTo(GlobalLock(*phGlobal));
			GlobalUnlock(*phGlobal);
		}
		else if(fmt->cfFormat == CF_HDROP)
		{
			*phGlobal = m_pShellItemList->CreateHDROP();
			if(!*phGlobal)
				return E_FAIL;
		}
		else
		{
			ATLTRACE("CDataObject_ShellItemList::GetData(0x%X) DV_E_FORMATETC\n", fmt->cfFormat);
			return DV_E_FORMATETC;
		}
		return S_OK;
	}
};

/*
inline bool MtlCopyFile(const CString& strTo, const CSimpleArray<CString>& arrFiles)
{
	LPTSTR lpszFrom = _MtlMakeFileOperationBuffer(arrFiles);

	CString strTo_ = strTo + _T('\0');
	SHFILEOPSTRUCT fop = { NULL, FO_COPY, lpszFrom, strTo_, FOF_ALLOWUNDO, FALSE, NULL, NULL };

	bool bOk = (::SHFileOperation(&fop) == 0);

	delete [] lpszFrom;
	return bOk;
}

inline bool MtlMoveFile(const CString& strTo, const CSimpleArray<CString>& arrFiles)
{
	LPTSTR lpszFrom = _MtlMakeFileOperationBuffer(arrFiles);

	CString strTo_ = strTo + _T('\0');
	SHFILEOPSTRUCT fop = { NULL, FO_MOVE, lpszFrom, strTo_, FOF_ALLOWUNDO, FALSE, NULL, NULL };

	bool bOk = (::SHFileOperation(&fop) == 0);

	delete [] lpszFrom;
	return bOk;
}
*/

} // namespace mew
