// TabbedContainer.h

#pragma once

#ifndef __cplusplus
	#error TabbedContainer.h requires C++ compilation
#endif

#ifndef __ATLAPP_H__
	#error TabbedContainer.h requires atlapp.h to be included first
#endif

#ifndef __ATLWIN_H__
	#error TabbedContainer.h requires atlwin.h to be included first
#endif

#ifndef __ATLFRAME_H__
	#error TabbedContainer.h requires atlframe.h to be included first
#endif

#ifndef __CUSTOMTABCTRL_H__
	#error TabbedContainer.h requires CustomTabCtrl.h to be included first
#endif

/////////////////////////////////////////////////////////////////////////////
//
// CCustomTabOwnerImpl
//  an MI template to help implement the owner window that uses CustomTabCtrl
//  to switch between windows / views
//
/////////////////////////////////////////////////////////////////////////////

template <class T, class TTabCtrl>
class CCustomTabOwnerImpl
{
public:
	typedef typename TTabCtrl::ItemType	ItemType;
// Member variables
private:
	TTabCtrl m_TabCtrl;
	CSize m_IconSize;
	int m_nTabAreaHeight;

// Constructors
public:
	CCustomTabOwnerImpl() :
		m_IconSize(16, 16),
		m_nTabAreaHeight(24)
	{
	}

// Overrideables
public:
	// A derived class might not need to override this although they can.
	// (but they will probably need to specialize SetTabAreaHeight)
	int CalcTabAreaHeight(void)
	{
		// Dynamically figure out a reasonable tab area height
		// based on the tab's font metrics

		const int nNominalHeight = max(GetIconSize().cy + 2, 24);
		const int nNominalFontLogicalUnits = 11;	// 8 point Tahoma with 96 DPI

		// Initialize nFontLogicalUnits to the typical case
		// appropriate for CDotNetTabCtrl
		LOGFONT lfIcon = { 0 };
		::SystemParametersInfo(SPI_GETICONTITLELOGFONT, sizeof(lfIcon), &lfIcon, 0);
		int nFontLogicalUnits = -lfIcon.lfHeight;

		// Use the actual font of the tab control
		TTabCtrl& TabCtrl = this->GetTabCtrl();
		if(TabCtrl.IsWindow())
		{
			HFONT hFont = TabCtrl.GetFont();
			if(hFont != NULL)
			{
				CDC dc = TabCtrl.GetDC();
				CFontHandle hFontOld = dc.SelectFont(hFont);
				TEXTMETRIC tm = {0};
				dc.GetTextMetrics(&tm);
				nFontLogicalUnits = tm.tmAscent;
				dc.SelectFont(hFontOld);
			}
		}

		return nNominalHeight + ( ::MulDiv(nNominalHeight, nFontLogicalUnits, nNominalFontLogicalUnits) - nNominalHeight ) / 2;
	}

// Methods
public:
	TTabCtrl& GetTabCtrl(void)
	{
		return m_TabCtrl;
	}

	CSize GetIconSize() const
	{
		return m_IconSize;
	}
	void SetIconSize(SIZE sz)
	{
		m_IconSize = sz;
	}
	int GetTabAreaHeight(void) const
	{
		return m_nTabAreaHeight;
	}
	void SetTabAreaHeight(int nNewTabAreaHeight)
	{
		m_nTabAreaHeight = nNewTabAreaHeight;
	}

	//
	// ^uւ̃ANZX
	//

	int SetCurSel(int index)
	{
		return m_TabCtrl.SetCurSel(index);
	}

	ItemType* GetItem(HWND hWnd) const
	{
		int index = GetTab(hWnd);
		return index < 0 ? NULL : m_TabCtrl.GetItem(index);
	}

	ItemType* GetItem(int index) const
	{
		return m_TabCtrl.GetItem(index);
	}

	int GetTab(HWND hWnd) const
	{
		int nCount = m_TabCtrl.GetItemCount();
		for(int i = 0; i < nCount; i++)
		{
			if(GetItem(i)->GetTabView() == hWnd)
				return i;
		}
		return -1;
	}

	int AddTab(HWND hWnd, LPCTSTR sTabText, LPCTSTR sToolTipText, HICON hIcon, int index = -1)
	{
		if(hWnd == NULL)
		{
			return -1;
		}

		ItemType* pItem = m_TabCtrl.AllocItem();
		if(pItem)
		{
			pItem->SetText(sTabText);
			if(sToolTipText)
				pItem->SetToolTip(sToolTipText);
			pItem->SetIcon(hIcon);
			pItem->SetIconSize(GetIconSize());
			pItem->SetTabView(hWnd);
			if(index < 0) index = m_TabCtrl.GetItemCount();
			return m_TabCtrl.InsertItem(index, pItem);
		}

		return -1;
	}

	BOOL RemoveTab(HWND hWnd)
	{
		int nTab = GetTab(hWnd);
		if(nTab < 0)
			return FALSE;
		return m_TabCtrl.DeleteItem(nTab);
	}
};

/////////////////////////////////////////////////////////////////////////////
//
// CTabbedContainerImpl
//
/////////////////////////////////////////////////////////////////////////////

template <
	class T,
	class TTabCtrl = CDotNetTabCtrl<CDotNetTabItem>,
	class TBase = CWindow,
	class TWinTraits = CControlWinTraits >
class CTabbedContainerImpl :
	public CWindowImpl<T, TBase, TWinTraits >,
	public CCustomTabOwnerImpl< T, TTabCtrl >
{
protected:
	typedef CTabbedContainerImpl< T, TTabCtrl, TBase, TWinTraits > thisClass;

// Member variables
protected:
	HWND m_hWndActive;

// Constructors
public:
	CTabbedContainerImpl() : m_hWndActive(NULL)
	{
	}

// Operation
	int GetActiveTab() const
	{
		return GetTab(m_hWndActive);
//		return m_TabCtrl.GetCurSel();
	}

	HWND GetActiveView() const
	{
		return m_hWndActive;
	}

// Message Handling
public:
	DECLARE_WND_CLASS_EX(_T("TabbedContainer"), 0, COLOR_APPWORKSPACE)

	BEGIN_MSG_MAP(thisClass)
		MESSAGE_HANDLER(WM_CREATE		, OnCreate)
		MESSAGE_HANDLER(WM_SETTINGCHANGE, OnSettingChange)
		MESSAGE_HANDLER(WM_ERASEBKGND	, OnEraseBkgnd)
		MESSAGE_HANDLER(WM_SETFOCUS		, OnSetFocus)
		MESSAGE_HANDLER(WM_SIZE			, OnSize)
		MESSAGE_HANDLER(WM_FORWARDMSG	, OnForwardMsg)
		MESSAGE_HANDLER(WM_PARENTNOTIFY	, OnParentNotify)

		NOTIFY_CODE_HANDLER(CTCN_DELETEITEM	, OnDeleteItem)
		NOTIFY_CODE_HANDLER(CTCN_SELCHANGING, OnSelChanging)
		NOTIFY_CODE_HANDLER(CTCN_SELCHANGE	, OnSelChange)

		REFLECT_NOTIFICATIONS()
	END_MSG_MAP()

	LRESULT OnCreate(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		// "baseClass::OnCreate()"
		LRESULT lRet = DefWindowProc(uMsg, wParam, lParam);
		bHandled = TRUE;

		GetTabCtrl().Create(m_hWnd);

		T* pT = static_cast<T*>(this);
		pT->SetTabAreaHeight(pT->CalcTabAreaHeight());
		return 0;
	}

	LRESULT OnSettingChange(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		// Be sure tab gets message before we recalculate the tab area height,
		//  so that it can adjust its font metrics first.
		// NOTE: This causes the tab to get the WM_SETTINGCHANGE message twice,
		//  but that's OK.
		GetTabCtrl().SendMessage(uMsg, wParam, lParam);

		T* pT = static_cast<T*>(this);
		pT->SetTabAreaHeight(pT->CalcTabAreaHeight());

		bHandled = FALSE;
		return 0;
	}

	LRESULT OnEraseBkgnd(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/)
	{
		if(GetTabCtrl().GetItemCount() > 0)
		{	// r[łꍇ́AANeBuȃr[SʂXV邽ߔwi`悷KvB
			return 1;
		}
		else
		{	// r[ꍇ́Awi`悷B
			return DefWindowProc();
		}
	}

	LRESULT OnSetFocus(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		// NOTE: ::IsWindowVisible(m_hWndActive) will be false if
		//  the frame is maximized.  So just use "IsWindow" instead.
		if(m_hWndActive != NULL && ::IsWindow(m_hWndActive))
		{
			::SetFocus(m_hWndActive);
		}

		bHandled = FALSE;
		return 1;
	}

	LRESULT OnSize(UINT /*uMsg*/, WPARAM wParam, LPARAM /*lParam*/, BOOL& bHandled)
	{
		if(wParam != SIZE_MINIMIZED)
		{
			T* pT = static_cast<T*>(this);
			pT->UpdateLayout();
		}
		bHandled = FALSE;
		return 1;
	}

	LRESULT OnForwardMsg(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& /*bHandled*/)
	{
		return ::SendMessage(m_hWndActive, WM_FORWARDMSG, wParam, lParam);
	}

	LRESULT OnParentNotify(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		UINT what = LOWORD(wParam);
		if(what == WM_DESTROY)
		{
			// r[jꂽꍇɃ^u폜B
			HWND hChild = (HWND)lParam;
			if(::GetParent(hChild) == m_hWnd && hChild != GetTabCtrl())
			{	// ̎qłA^uRg[ŖȂ...
				RemoveTab(hChild);
			}
		}
		bHandled = FALSE;
		return 0;
	}

	LRESULT OnDeleteItem(int /*idCtrl*/, LPNMHDR pnmh, BOOL& bHandled)
	{
		NMCTCITEM* item = (NMCTCITEM*)pnmh;
		ItemType* pItem = GetTabCtrl().GetItem(item->iItem);
		if(m_hWndActive == pItem->GetTabView())
			m_hWndActive = NULL;
		bHandled = FALSE;
		return 0;
	}

	LRESULT OnSelChanging(int /*idCtrl*/, LPNMHDR /*pnmh*/, BOOL& bHandled)
	{
		bHandled = FALSE;
		return 0;
	}

	LRESULT OnSelChange(int /*idCtrl*/, LPNMHDR /*pnmh*/, BOOL& bHandled)
	{
		INT nCount = GetTabCtrl().GetItemCount();
		for(INT i = 0; i < nCount; i++)
		{
			if(GetTabCtrl().GetSel(i))
			{
				ItemType* pItem = GetTabCtrl().GetItem(i);
				if(pItem->UsingTabView())
				{
					HWND hWndNew = pItem->GetTabView();
					HWND hWndOld = m_hWndActive;
					if( hWndNew != hWndOld )
					{
						m_hWndActive = hWndNew;

						//updatelayout is going to essentially do a
						//  "ShowWindow(hWndNew, SW_SHOW)" for us
						// (Call the most derived class's version of UpdateLayout)
						T* pT = static_cast<T*>(this);
						pT->UpdateLayout();

						if(hWndOld)
						{
							::ShowWindow(hWndOld, SW_HIDE);
						}

						HWND hFocus = ::GetFocus();
						if(m_hWnd == hFocus || IsChild(hFocus))
							::SetFocus(hWndNew);
					}
				}
				bHandled = FALSE;
				return 0;
			}
		}
		// ^u
		if(m_hWndActive)
		{
			::ShowWindow(m_hWndActive, SW_HIDE);
			m_hWndActive = NULL;
		}
		bHandled = FALSE;
		return 0;
	}

// Overrides from CCustomTabOwnerImpl
public:
	void SetTabAreaHeight(int nNewTabAreaHeight)
	{
		__super::SetTabAreaHeight(nNewTabAreaHeight);
		T* pT = static_cast<T*>(this);
		pT->UpdateLayout();
		Invalidate();
	}

public:
	void UpdateLayout()
	{
		CRect rect;
		GetClientRect(&rect);

		int nWindowPosCount=0;
		if(GetTabCtrl()) nWindowPosCount++;
		if(m_hWndActive) nWindowPosCount++;

		if(nWindowPosCount > 0)
		{
			HDWP hdwp = BeginDeferWindowPos(nWindowPosCount);
			DWORD dwStyle = (DWORD)GetTabCtrl().GetWindowLong(GWL_STYLE);
			if(CTCS_BOTTOM == (dwStyle & CTCS_BOTTOM))
			{
				if(GetTabCtrl())
				{
					hdwp = ::DeferWindowPos(
						hdwp,
						GetTabCtrl(),
						NULL,
						rect.left, rect.bottom - GetTabAreaHeight(),
						rect.Width(), GetTabAreaHeight(),
						SWP_NOZORDER | SWP_NOACTIVATE);
				}
				if(m_hWndActive)
				{
					hdwp = ::DeferWindowPos(
						hdwp,
						m_hWndActive,
						NULL,
						rect.left, rect.top,
						rect.Width(), rect.Height() - GetTabAreaHeight(),
						SWP_NOZORDER | SWP_SHOWWINDOW);
				}
			}
			else
			{
				if(GetTabCtrl())
				{
					hdwp = ::DeferWindowPos(
						hdwp,
						GetTabCtrl(),
						NULL,
						rect.left, rect.top,
						rect.Width(), GetTabAreaHeight(),
						SWP_NOZORDER | SWP_NOACTIVATE);
				}
				if(m_hWndActive)
				{
					hdwp = ::DeferWindowPos(
						hdwp,
						m_hWndActive,
						NULL,
						rect.left, rect.top + GetTabAreaHeight(),
						rect.Width(), rect.Height() - GetTabAreaHeight(),
						SWP_NOZORDER | SWP_SHOWWINDOW);
				}
			}
			EndDeferWindowPos(hdwp);
		}

		GetTabCtrl().UpdateLayout();
		GetTabCtrl().Invalidate();
	}
};
